/*
  BlueNIC - Bluetooth PAN profile implementation for BlueZ
  Copyright (C) 2002 Sony Corporation

  Author: Johannes Loebbert <loebbert@sony.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2, as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/

/* 
  ChangeLog:
  2001/04/17 created by Johannes Loebbert
*/

/*
 * $Id: bnep_core.c,v 1.13 2002/07/08 14:28:08 johanloe Exp $
*/

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/module.h>

#include <linux/slab.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <net/bluetooth/bluetooth.h>
#include <net/bluetooth/l2cap.h>

#include "bnep_core.h"

#ifndef CONFIG_BNEP_DEBUG
#undef  BT_DBG
#define BT_DBG( A... )
#endif

//Counter for loaded eth-interfaces
static int bnep_dev_counter;
static struct bnep_owner_struct bnep_owner;

/***************************************************************************/
#ifdef BNEP_TEST
int test_ignore_connect;
int test_ignore_netfilter;
int test_ignore_multicastfilter;
#endif

struct semaphore bnep_mc_send_sem;
kthread_t bnep_mc_send_thread;

/**********************************************************************/
struct net_device *root_dev;

/**********************************************************************/
#ifdef CONFIG_BLUEZ_DEBUG
static unsigned long rx_counter = 0;
static unsigned long tx_counter = 0;
#endif

/***************************************************************************/
struct bnep_connection *bnep_get_conn_by_bdaddr(struct net_device *dev, bdaddr_t * target_addr)
{
	struct list_head *ptr;
	struct bnep_connection *entry;
	struct bnep_connection *conn = NULL;
	bdaddr_t addr;
	struct bnep_private *priv = bnep_get_priv(dev);

	baswap(&addr, target_addr);

	for (ptr = priv->bnep_connections.next; ptr != &priv->bnep_connections; ptr = ptr->next) {
		entry = list_entry(ptr, struct bnep_connection, list);
		if (bacmp(&bluez_pi(entry->sk->sk)->dst, &addr) == 0) {
			conn = entry;
			break;
		}
	}

	if (conn == NULL) {
		BT_DBG("connection %s not found in connection table", batostr(target_addr));
	}

	return conn;
}


/***************************************************************************/
static int bnep_rx_multicast_filter_rsp(struct bnep_connection *bnep_conn,
					struct sk_buff *skb_input)
{
	__u16 ResponseMessage;

	//Copy response code and remove info from header
	if (skb_input->len < 2) {	
		BT_ERR("packet to short => discard");
		return 1;
	}
	memcpy(&ResponseMessage, skb_input->data, 2);
	skb_pull(skb_input, 2);
	
	BT_DBG("multicast set control rsp: 0x%.4x", bnep16_to_cpu(ResponseMessage));

	if (bnep_conn->filter_request.req_status == BNEP_REQ_PENDING) {
		bnep_conn->filter_request.req_status = BNEP_REQ_DONE;
		bnep_conn->filter_request.req_result = ResponseMessage;
		wake_up_interruptible(&(bnep_conn->filter_request.req_wait_q));
	} else {
		BT_DBG("connection not waiting for filter response => discard");
	}

	return 0;
}

/***************************************************************************/
static int bnep_rx_net_filter_rsp(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{
	__u16 ResponseMessage;

	//Copy response code and remove info from header
	if (skb_input->len < 2) {	
		BT_ERR("packet to short => discard");
		return 1;
	}
	memcpy(&ResponseMessage, skb_input->data, 2);
	skb_pull(skb_input, 1);

	BT_DBG("protocol filter set control rsp: 0x%.4x", ResponseMessage);

	if (bnep_conn->filter_request.req_status == BNEP_REQ_PENDING) {
		bnep_conn->filter_request.req_status = BNEP_REQ_DONE;
		bnep_conn->filter_request.req_result = ResponseMessage;
		wake_up_interruptible(&(bnep_conn->filter_request.req_wait_q));
	} else {
		BT_DBG("connection not waiting for filter response => discard");
	}
	return 0;
}

/***************************************************************************/
static int bnep_rx_unknown_controldata(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{
	__u8 unknown_command;

	//Copy control type and remove info from header
	if (skb_input->len < 1) {	
		BT_ERR("packet to short => discard");
		return 1;
	}
	memcpy(&unknown_command, skb_input->data, 1);
	skb_pull(skb_input, 1);

	BT_DBG("unknown command: 0x%x", unknown_command);

	return 0;
}

/***************************************************************************/
static int bnep_tx_unknown_controldata(struct bnep_connection *bnep_conn, 
			__u8 unknown_control_code)
{
	struct sk_buff *skb;
	int packet_size = 3;

	//Create control packet
	if (!(skb = alloc_skb(packet_size, GFP_ATOMIC)))
		return -ENOMEM;

	//Reserve space for control header 
	skb_reserve(skb, 2);

	//Insert unknown command
	memcpy(skb_put(skb, 1), &unknown_control_code, 1);

	//Add bnep header and send message    
	bnep_tx_controldata(bnep_conn, BNEP_CONTROL_COMMAND_NOT_UNDERSTOOD, 0, skb);

	return 0;
}

/***************************************************************************/
/***************************************************************************/
int bnep_tx_controldata(struct bnep_connection *bnep_conn, __u8 ControlType, int ExtensionFlag,
			struct sk_buff *skb)
{
	if (!bnep_conn) {
		BT_DBG("Attempt to send on non existing connection");
		dev_kfree_skb(skb);
		return 1;
	}

	if (bnep_conn -> status == BNEP_CONN_CLOSED && 
		ControlType != BNEP_CONTROL_COMMAND_NOT_UNDERSTOOD &&
		ControlType != BNEP_SETUP_CONNECTION_RESPONSE_MSG &&
		ControlType != BNEP_SETUP_CONNECTION_REQUEST_MSG) {
		BT_DBG("Attempt to send to closed connection");
		dev_kfree_skb(skb);
		return 1;
	}

	if (skb_headroom(skb) < 2) {
		//Not enough headroom for the two bytes for a control packet
		BT_ERR("Not enough headspace to insert control packet header");
		dev_kfree_skb(skb);
	} else {
		__u8 bnep_header = BNEP_CONTROL;
		__u8 control_type = ControlType;

		if (ExtensionFlag) {
			BT_DBG("set extension flag");
			bnep_header = bnep_header | 0X80;
		}
		//Add control type
		memcpy(skb_push(skb, 1), &control_type, 1);

		//Add bnep type with extension flag
		memcpy(skb_push(skb, 1), &bnep_header, 1);

		skb_queue_tail(&bnep_conn->sk->sk->write_queue, skb);
		wake_up_interruptible(bnep_conn->sk->sk->sleep);
	}
	return 0;
}

/***************************************************************************/
static int bnep_tx_etherdata(struct bnep_connection *bnep_conn, struct bnep_connection *entry,
			       struct sk_buff *skb_input, struct ethhdr *eth, __u16 * h802_1p,
			       int ExtensionFlag, __u8 packet_type)
{
	struct sk_buff *skb = NULL;
	__u16 proto;
	__u8 bnep_header;
	bdaddr_t tmp_addr;
	__u8 header_size = 1 + 2 * 6 + 2;

	if (!entry || entry -> status == BNEP_CONN_CLOSED) {
		BT_DBG("Attempt to send on non existing or closed connection");
		return 1;
	}

	if (entry)
		baswap(&tmp_addr, &bluez_pi(entry->sk->sk)->dst);

	if (h802_1p) {
		//Increase header size for 802.1p header
		header_size += 4;
	}
	//Prepare packet
	if (!(skb = alloc_skb(skb_input->len + header_size, GFP_ATOMIC)))
		return -ENOMEM;
	skb_reserve(skb, header_size);

	//Copy packet data
	memcpy(skb_put(skb, skb_input->len), skb_input->data, skb_input->len);

	//Insert packet type
	if (h802_1p) {
		//Insert real protocol
		proto = (__u16) eth->h_proto;
		memcpy(skb_push(skb, 2), &proto, 2);
		//Insert 802.1p header
		memcpy(skb_push(skb, 2), &h802_1p, 2);
		//Insert 801.p protocol
		proto = 0x0081;
		memcpy(skb_push(skb, 2), &proto, 2);
	} else {
		proto = (__u16) eth->h_proto;
		memcpy(skb_push(skb, 2), &proto, 2);
	}

	if (packet_type == BNEP_PACKET_TYPE_AUTO) {
		//Automatic compression detection
		if (bnep_conn == NULL) {
			//Packet was created by local node
			bdaddr_t bd_source_addr;
			baswap(&bd_source_addr, &bluez_pi(entry->sk->sk)->src);
			if (memcmp(eth->h_source, &bd_source_addr, ETH_ALEN) != 0) {
				//Issued from the bridge code => must be general ethernet
				packet_type = BNEP_GENERAL_ETHERNET;
			} else if (memcmp(eth->h_dest, &tmp_addr, ETH_ALEN) == 0) {
				//Local issued packet to next hop => compressed
				packet_type = BNEP_COMPRESSED_ETHERNET;
			} else {
				//Local issued packet => Dest. only
				packet_type = BNEP_COMPRESSED_ETHERNET_DEST_ONLY;
			}
		} else {
			//Packet forwarding code
			if (memcmp(eth->h_dest, &tmp_addr, ETH_ALEN) == 0) {
				//Delivering to final destination => GeneralEthernet Source Only
				packet_type = BNEP_COMPRESSED_ETHERNET_SOURCE_ONLY;
			} else {
				//Delivered to general node
				packet_type = BNEP_GENERAL_ETHERNET;
			}
		}
	}

	switch (packet_type) {
	case BNEP_GENERAL_ETHERNET:
		BT_DBG("packet type: General Ethernet");
		memcpy(skb_push(skb, 6), eth->h_source, 6);
		memcpy(skb_push(skb, 6), eth->h_dest, 6);
		bnep_header = BNEP_GENERAL_ETHERNET;
		break;
	case BNEP_COMPRESSED_ETHERNET:
		BT_DBG("packet type: Compressed Ethernet");
		bnep_header = BNEP_COMPRESSED_ETHERNET;
		break;
	case BNEP_COMPRESSED_ETHERNET_SOURCE_ONLY:
		BT_DBG("packet type: Compressed Ethernet Source only");
		memcpy(skb_push(skb, 6), eth->h_source, ETH_ALEN);
		bnep_header = BNEP_COMPRESSED_ETHERNET_SOURCE_ONLY;
		break;
	case BNEP_COMPRESSED_ETHERNET_DEST_ONLY:
		BT_DBG("packet type: Compressed Ethernet Destination only");
		memcpy(skb_push(skb, 6), eth->h_dest, ETH_ALEN);
		bnep_header = BNEP_COMPRESSED_ETHERNET_DEST_ONLY;
		break;
	}

	//Modify for extension flag
	if (ExtensionFlag)
		bnep_header = bnep_header | 0x80;

	//Adding BNEP header
	memcpy(skb_push(skb, 1), &bnep_header, 1);

	if (entry->sk) {
		BT_DBG("Source address: %s", batostr((bdaddr_t *) eth->h_source));
		BT_DBG("Dst. address:   %s", batostr((bdaddr_t *) eth->h_dest));
		BT_DBG("Protocol:       0x%x %s", bnep16_to_cpu(eth->h_proto),
				(h802_1p) ? "(802.1p)" : "");

		//memcpy(skb -> cb , &entry -> sk , sizeof(struct socket*));
		skb_queue_tail(&entry->sk->sk->write_queue, skb);
		wake_up_interruptible(entry->sk->sk->sleep);
	} else {
		BT_DBG("transmission error. SK not connected");
		kfree_skb(skb);
	}

	return 0;
}

/***************************************************************************/
static int bnep_tx_apply_multicast_filter(struct bnep_connection *bnep_conn, struct ethhdr *eth)
{
	int result = 0;

	if (bnep_conn->filter_info.MulticastFilter) {
		int counter = 0;
		__u64 addr = 0;
		result = 1;

		for (counter = 0; counter < 6; counter++)
			((__u8 *) & addr)[counter] = eth->h_dest[5 - counter];

		for (counter = 0; counter < bnep_conn->filter_info.multicast_filter_count;
		     counter++) {
			BT_DBG("Filter %d: Value: 0x%.4x%.8x Filter: 0x%.4x%.8x - 0x%.4x%.8x", counter,
				((__u32 *) & addr)[1], ((__u32 *) & addr)[0],
				((__u32 *) & bnep_conn->filter_info.MulticastFilter[counter].start)[1],
				((__u32 *) & bnep_conn->filter_info.MulticastFilter[counter].start)[0],
				((__u32 *) & bnep_conn->filter_info.MulticastFilter[counter].stop)[1],
				((__u32 *) & bnep_conn->filter_info.MulticastFilter[counter].stop)[0]);

			if (bnep_conn->filter_info.MulticastFilter[counter].start <= addr) {
				if (addr <= bnep_conn->filter_info.MulticastFilter[counter].stop)
					return 0;
			}
		}
	}
	return result;
}

/***************************************************************************/
static int bnep_tx_apply_protocol_filter(struct bnep_connection *bnep_conn, struct ethhdr *eth)
{
	int result = 0;

	if (bnep_conn->filter_info.NetFilter) {
		int counter = 0;
		__u16 protocol = bnep16_to_cpu(eth->h_proto);
		result = 1;

		for (counter = 0; counter < bnep_conn->filter_info.net_filter_count; counter++) {
			BT_DBG("Filter %d: 0x%.4x - 0x%.4x Value: 0x%.4x", counter,
					bnep_conn->filter_info.NetFilter[counter].start,
					bnep_conn->filter_info.NetFilter[counter].stop, protocol);
			if (bnep_conn->filter_info.NetFilter[counter].start <= protocol
			    && protocol <= bnep_conn->filter_info.NetFilter[counter].stop)
				return 0;
		}
	}
	return result;
}

/***************************************************************************/
//Processing packet for BT domain
int bnep_tx_etherdata_multiplexor(struct net_device *dev, struct bnep_connection *bnep_conn,
			       struct sk_buff *skb_input, struct ethhdr *eth, __u16 *h802_1p,
			       int ExtensionFlag, __u8 packet_type)
{
	//eth is just a reference don't free
	struct list_head *ptr;
	struct bnep_connection *entry;
	struct bnep_private *priv;
	bdaddr_t tmp_addr;
	int send_packet;

	if (dev == NULL) {
		BT_DBG("can not send to unknown device");
		return 0;
	}

	priv = bnep_get_priv(dev);

	for (ptr = priv->bnep_connections.next; ptr != &priv->bnep_connections; ptr = ptr->next) {
		entry = list_entry(ptr, struct bnep_connection, list);
		baswap(&tmp_addr, &bluez_pi(entry->sk->sk)->dst);

		//Do not send back to incoming connection
		if (entry == bnep_conn)
			continue;

		//Initial setting
		send_packet = 0;

		//Apply bridging filter
		if (priv->connection_counter == 1) {
			//Slave => Default GW || Master -> Only one slave
			send_packet = 1;
		} else if (*eth->h_dest & 1 || memcmp(eth->h_dest, &tmp_addr, ETH_ALEN) == 0) {
			//Multi-Broadcast or unicast to directly connected node
			send_packet = 1;
		}

		if (send_packet) {
			//Packet type filter
			send_packet = bnep_tx_apply_protocol_filter(entry, eth) ? 0 : 1;
			if (send_packet && *eth->h_dest & 1) {
				//Multicast filter
				send_packet = bnep_tx_apply_multicast_filter(entry, eth) ? 0 : 1;
			}
		}

		if (send_packet)
			bnep_tx_etherdata(bnep_conn, entry, skb_input, eth, h802_1p,
					ExtensionFlag, packet_type);
		else {
			BT_DBG("packet will not be send to %s "
				"due to filter/bridging settings", batostr(&tmp_addr));
		}
	}

	if (h802_1p)
		kfree(h802_1p);

	kfree_skb(skb_input);
	return 0;
}

/***************************************************************************/
static int bnep_rx_etherdata(struct bnep_connection *bnep_conn, struct sk_buff *skb_input,
			     int PacketType, int ExtensionFlag)
{
	struct ethhdr eth;
	int ExtensionHeadersLength = 0;
	__u16 *h802_1p = NULL;
	struct bnep_private *priv = bnep_get_priv(bnep_conn->net_dev);

	if (bnep_conn -> status == BNEP_CONN_CLOSED) {
		BT_DBG("connection not ready => drop data packet");
		kfree_skb(skb_input);
		return 1;	
	}

	//Copy bnep address information in ethernet struct 
	switch (PacketType) {
	case BNEP_GENERAL_ETHERNET:{
			BT_DBG("Packet type: General Ethernet");
			if (skb_input->len < 2 * ETH_ALEN) goto packet_length_error; 	
			memcpy(&(eth.h_dest), skb_input->data, ETH_ALEN);
			memcpy(&(eth.h_source), &(skb_input->data)[ETH_ALEN], ETH_ALEN);
			skb_pull(skb_input, 2 * ETH_ALEN);
			break;
		}
	case BNEP_COMPRESSED_ETHERNET:{
			bdaddr_t tmp_addr;
			BT_DBG("Packet type: Compressed Ethernet");
			memcpy(&(eth.h_dest), &priv->source_address, ETH_ALEN);
			baswap(&tmp_addr, &bluez_pi(bnep_conn->sk->sk)->dst);
			memcpy(&(eth.h_source), &tmp_addr, ETH_ALEN);
			break;
		}
	case BNEP_COMPRESSED_ETHERNET_SOURCE_ONLY:{
			BT_DBG("Packet type: Compressed Ethernet Source only");
			if (skb_input->len < ETH_ALEN) goto packet_length_error; 	
			memcpy(&(eth.h_dest), &priv->source_address, ETH_ALEN);
			memcpy(&(eth.h_source), skb_input->data, ETH_ALEN);
			skb_pull(skb_input, ETH_ALEN);
			break;
		}
	case BNEP_COMPRESSED_ETHERNET_DEST_ONLY:{
			bdaddr_t tmp_addr;
			BT_DBG("Packet type: Compressed Ethernet Destination only");
			if (skb_input->len < ETH_ALEN) goto packet_length_error;	
			memcpy(&(eth.h_dest), skb_input->data, ETH_ALEN);
			skb_pull(skb_input, ETH_ALEN);
			baswap(&tmp_addr, &bluez_pi(bnep_conn->sk->sk)->dst);
			memcpy(&(eth.h_source), &tmp_addr, ETH_ALEN);
			break;
		}
	default : {
			BT_DBG("Packet type: Unknown");
			kfree_skb(skb_input);
			return 1;	
		}
	}

	//Copy protocol from BNEP header
	if (skb_input->len < 2) goto packet_length_error; 	
	memcpy(&(eth.h_proto), skb_input->data, 2);
	skb_pull(skb_input, 2);

	BT_DBG("Source address: %s", batostr((bdaddr_t *) eth.h_source));
	BT_DBG("Dst. address: %s", batostr((bdaddr_t *) eth.h_dest));

	//Handle extension header if available and calculate extensions lenght 
	BT_DBG("ExtensionHeader: %s", ExtensionFlag ? "Yes" : "No");
	if (ExtensionFlag) {
		//Evaluate extension header
		if (bnep_rx_extension_header(bnep_conn , skb_input, &ExtensionHeadersLength)) {
			BT_ERR("error in extension header");
			goto packet_length_error;
		}
	}

	//802.1p handling                                             
	if (bnep16_to_cpu(eth.h_proto) == 0x8100) {
		if (skb_input->len < ExtensionHeadersLength + 2) goto packet_length_error; 	
		h802_1p = (__u16 *) kmalloc(2, GFP_ATOMIC);
		//Remove 801.1p header
		memcpy(h802_1p, &skb_input->data[ExtensionHeadersLength], 2);
		ExtensionHeadersLength += 2;
		//Copy real protocol from BNEP header
		memcpy(&(eth.h_proto), &skb_input->data[ExtensionHeadersLength], 2);
		ExtensionHeadersLength += 2;
		BT_DBG("Protocol: 0x%x (802.1p)", bnep16_to_cpu(eth.h_proto));
	} else {
		BT_DBG("Protocol: 0x%x", bnep16_to_cpu(eth.h_proto));
	}

	if (skb_input->len < ExtensionHeadersLength) 
		goto packet_length_error;
	else {
		//Send data payload package to own protocol stack
		struct sk_buff *skb;
		int PayloadLength = skb_input->len - ExtensionHeadersLength;

		//Mem alloc check
		if (!(skb = alloc_skb(PayloadLength + ETH_HLEN + (h802_1p ? 4 : 0) + 2, GFP_ATOMIC)))
			return -ENOMEM;

		//Set package header/tail - room in target frame 
		//(+2 => allignement of ip header with byte border)
		skb_reserve(skb, ETH_HLEN + 2 + (h802_1p ? 4 : 0));

		//Copy data and header to new frame
		memcpy(skb_put(skb, PayloadLength), &(skb_input->data[ExtensionHeadersLength]),
		       PayloadLength);

		if (h802_1p) {
			//Restore 802.1q header
			memcpy(skb_push(skb , 2) , &eth.h_proto , 2);
			eth.h_proto = 0x0081;
			memcpy(skb_push(skb , 2) , h802_1p , 2);
		}		       
		
		memcpy(skb_push(skb, ETH_HLEN), &eth, ETH_HLEN);

		//Sending ethernet packet to device driver layer
		device_rx_data(skb, bnep_conn->net_dev);
	}

	if (PacketType == BNEP_GENERAL_ETHERNET || PacketType == BNEP_COMPRESSED_ETHERNET_DEST_ONLY) {
		//forward only ethernet and dest_only packets
		if (priv->connection_counter > 1) {
			//Forward packet to other nodes
			bnep_tx_etherdata_multiplexor(bnep_conn->net_dev, bnep_conn, skb_input, &eth,
						   h802_1p, ExtensionFlag, BNEP_PACKET_TYPE_AUTO);
		} else 
			kfree_skb(skb_input);
	} else {
		//Delete input packet
		kfree_skb(skb_input);

		//Delete 802.1p header
		if (h802_1p)
			kfree(h802_1p);
	}

	return 0;

packet_length_error:
	BT_ERR("illegal packet length info in data packet or extension header");
	kfree_skb(skb_input);
	return 1;	
}

static int bnep_execute_controldata(struct bnep_connection *bnep_conn, struct sk_buff *skb_input ,
				int packet_length)
{
	__u8 ControlType;

	//Copy control type and remove info from header
	if (skb_input -> len < 1) goto packet_length_error;
	memcpy(&ControlType, skb_input->data, 1);
	skb_pull(skb_input, 1);
	packet_length--;

	//Process control packet
	BT_DBG("BNEP Control type 0x%.2x", ControlType);
	switch (ControlType) {
	case BNEP_SETUP_CONNECTION_REQUEST_MSG:
		return bnep_rx_connect_req(bnep_conn, skb_input);
		break;
		
	case BNEP_SETUP_CONNECTION_RESPONSE_MSG:
		return bnep_rx_connect_rsp(bnep_conn, skb_input);
		break;

	case BNEP_CONTROL_COMMAND_NOT_UNDERSTOOD:
		if (bnep_conn -> status == BNEP_CONN_CLOSED) goto connection_not_ready;
		return bnep_rx_unknown_controldata(bnep_conn, skb_input);
		break;

	case BNEP_FILTER_NET_TYPE_SET_MSG:
		if (bnep_conn -> status == BNEP_CONN_CLOSED) goto connection_not_ready;
		return bnep_rx_net_filter(bnep_conn, skb_input);
		break;
		
	case BNEP_FILTER_NET_TYPE_RESPONSE_MSG:
		if (bnep_conn -> status == BNEP_CONN_CLOSED) goto connection_not_ready;
		return bnep_rx_net_filter_rsp(bnep_conn, skb_input);
		break;

	case BNEP_FILTER_MULT_ADDR_SET_MSG:
		if (bnep_conn -> status == BNEP_CONN_CLOSED) goto connection_not_ready;
		return bnep_rx_multicast_filter(bnep_conn, skb_input);
		break;

	case BNEP_FILTER_MULT_ADDR_RESPONSE_MSG:
		if (bnep_conn -> status == BNEP_CONN_CLOSED) goto connection_not_ready;
		return bnep_rx_multicast_filter_rsp(bnep_conn, skb_input);
		break;

	default:
		//Send UNKNOWN COMMAND response packet
		bnep_tx_unknown_controldata(bnep_conn , ControlType);
		if (packet_length != -1 && packet_length < skb_input->len) {
			//Remove remainings of packet if length is known
			BT_DBG("unknown control ignore 0x%x bytes", packet_length);
			skb_pull(skb_input , packet_length);
		}
		else 
			goto packet_length_error;
		break;
	}
	
	return 0;


connection_not_ready:
	BT_DBG("connection not ready => drop data packet");
	return 0;
	
packet_length_error:	
	BT_ERR("illegal/unknwon control packet length info => skip remaining parts");
	return 1;

}

/******************************************************************************/
static int bnep_rx_controldata(struct bnep_connection *bnep_conn, struct sk_buff *skb_input,
				int ExtensionFlag)
{
	int ExtensionHeadersLength;
	int status = 0;

	//Execute command
	if (bnep_execute_controldata(bnep_conn , skb_input , -1)) {
		BT_ERR("error evaluating control packet => skip extension header");
		status = 1;	
	}
	else {
		BT_DBG("extension header: %s" , ExtensionFlag ? "yes" : "no");
		if (ExtensionFlag) {
			bnep_rx_extension_header(bnep_conn, skb_input , &ExtensionHeadersLength);
		}
	}

	//Delete input packet
	dev_kfree_skb(skb_input);
	return status;
}

/***************************************************************************/
static int bnep_rx_extension_header(struct bnep_connection *bnep_conn, struct sk_buff *skb_input,
				      int *ExtensionsLength)
{
	__u8 ExtensionType;
	int ExtensionFlag = 0;
	int error = 0;
	__u8 ExtensionPayloadLength;

	*ExtensionsLength = 0;

	//Evaluate all header and update action packet
	do {
		if (skb_input->len < 2) {
			//Packet to short for extension header => cancel evaluation
			BT_ERR("packet to short to contain extension");
			error = 1;
			break;
		}

		//Copy extension type[extension flag]
		memcpy(&ExtensionType, skb_input->data, 1);
		ExtensionFlag = ExtensionType & 0x80;

		//Copy extension length
		memcpy(&ExtensionPayloadLength, &skb_input->data[1], 1);

		if ((ExtensionType & 0x7F) == BNEP_EXTENSION_CONTROL) {

			BT_DBG("evaluate extension control header");
			
			//Remove type & length
			skb_pull(skb_input, 2);

			//Execute command in packet
			bnep_execute_controldata(bnep_conn , skb_input , ExtensionPayloadLength);
		} 
		else {
			//Ignore all extensions after first unknown extension was found
			BT_DBG("ignore unknown extension: type: 0x%x length: 0x%x (%d)",
				(ExtensionType & 0x7F), ExtensionPayloadLength, ExtensionPayloadLength);
			//Length of first header
			*ExtensionsLength = ExtensionPayloadLength + 2;
			//Iterate over remaining headers
			while (!error && ExtensionFlag) {
				if (*ExtensionsLength + 2 <= skb_input->len) {
					memcpy(&ExtensionType, &skb_input->data[*ExtensionsLength], 1);
					ExtensionFlag = ExtensionType & 0x80;
					memcpy(&ExtensionPayloadLength,
						&skb_input->data[*ExtensionsLength + 1], 1);
					BT_DBG("ignore unknown extension: type: 0x%x length: 0x%x (%d)",
						(ExtensionType & 0x7F), ExtensionPayloadLength , ExtensionPayloadLength);
					*ExtensionsLength += ExtensionPayloadLength + 2;
				} else {
					//Error: Packet is shorter than extension length => Cancel processing
					error = 1;
				}
			}
		}
	} while (ExtensionFlag && !error);

	return error;
}


/***************************************************************************/
static int bnep_rx_multicast_filter(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{
	struct multicast_range *new_filter = NULL;
	__u16 list_length;
	int valid_range = 1;
	
	struct sk_buff *skb;
	__u16 response_code = MCFILTER_OPERATION_SUCCESSFUL;
	int counter;

	//Copy list length and remove info from header
	if (skb_input->len < 2) {	
		BT_ERR("packet to short => discard");
		return 1;
	}
	memcpy(&list_length, skb_input->data, 2);
	list_length = bnep16_to_cpu(list_length);
	skb_pull(skb_input, 2);
	BT_DBG("List length:%.4x", list_length);

	if (list_length == 0) {
		BT_DBG("resetting filter");
		goto set_multicast_filter;
	} 
	
	if (list_length > skb_input->len) {
		BT_ERR("filter list lenght error: 0x%x > 0x%x => skip remaning packet",
				list_length, skb_input->len);
		return 1;
	}
	
	if (list_length % 6 != 0) {
		BT_ERR("filter list lenght error: odd number of addresses (0x%x) => skip filter",
				list_length);
		skb_pull(skb_input, list_length);
		return 0;
	}
	
#ifdef BNEP_TEST
		if (test_ignore_multicastfilter == TESTCASE_IGNORE_MESSAGE_ON) {
			BT_DBG("Ignore multicast test case activated. Response will not be sent");
		skb_pull(skb_input, list_length);
		return 0;
	}
#endif

	//Alloc space for new filter
	if (!(new_filter = (struct multicast_range *)
			kmalloc(sizeof(struct multicast_range[list_length / 12]), GFP_KERNEL))) {
		return -ENOMEM;
	}
	memset(new_filter , 0 , sizeof(struct multicast_range[list_length / 12]));

	//Copy values
	for (counter = 0; counter < list_length / 12; counter++) {
		memcpy(&new_filter[counter].start , skb_input->data, 6); 
		new_filter[counter].start = bnep64_to_cpu(new_filter[counter].start) >> 16;
		memcpy(&new_filter[counter].stop , &(skb_input->data)[6], 6); 
		new_filter[counter].stop = bnep64_to_cpu(new_filter[counter].stop) >> 16;
		skb_pull(skb_input, 12);

		valid_range = (new_filter[counter].start <= new_filter[counter].stop &&
					   new_filter[counter].start & 0x010000000000 && 
					   new_filter[counter].stop & 0x010000000000);

		BT_DBG("filter %d: 0x%.4x%.8x - 0x%.4x%.8x (%s)", counter,
				((__u32 *)&new_filter[counter].start)[1],
				((__u32 *)&new_filter[counter].start)[0],
				((__u32 *)&new_filter[counter].stop)[1],
				((__u32 *)&new_filter[counter].stop)[0] ,
				valid_range ? "valid" : "invalid");

				
		if (!valid_range) {
			response_code = cpu_to_bnep16(MCFILTER_FAILD_INVALID_MC_ADDR);
		}		
	}

set_multicast_filter:

	if (response_code == MCFILTER_OPERATION_SUCCESSFUL) {
		//Delete old filter settings
		if (bnep_conn->filter_info.MulticastFilter)
			kfree(bnep_conn->filter_info.MulticastFilter);

		//Set new filter
		bnep_conn->filter_info.MulticastFilter = new_filter;
		bnep_conn->filter_info.multicast_filter_count = list_length / 12;
	}

	//Create response control packet
	if (!(skb = alloc_skb(4, GFP_ATOMIC)))
		return -ENOMEM;
	skb_reserve(skb, 2);

	//Insert response code
	memcpy(skb_put(skb, 2), &response_code, 2);

	//Add bnep header and send message    
	bnep_tx_controldata(bnep_conn, BNEP_FILTER_MULT_ADDR_RESPONSE_MSG, 0, skb);

	return 0;
}
/***************************************************************************/
static int bnep_rx_net_filter(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{
	struct protocol_range *new_filter = NULL;
	__u16 list_length;

	struct sk_buff *skb;
	__u16 response_code = NETFILTER_OPERATION_SUCCESSFUL;
	int counter;

	//Copy list length and remove info from header
	if (skb_input->len < 2) {	
		BT_ERR("packet to short => discard");
		return 1;
	}
	memcpy(&list_length, skb_input->data, 2);
	list_length = bnep16_to_cpu(list_length);
	skb_pull(skb_input, 2);
	BT_DBG("List length:%.4x", list_length);

	if (list_length == 0) {
		BT_DBG("resetting filter");
		goto set_net_filter;
	} 
	
	if (list_length > skb_input->len) {
		BT_ERR("filter list lenght error: 0x%x > 0x%x => skip remaning packet",
				list_length, skb_input->len);
		return 1;
	}
	
	if (list_length % 2 != 0) {
		BT_ERR("filter list lenght error: odd number of bytes (0x%x) => skip filter",
				list_length);
		skb_pull(skb_input, list_length);
		return 0;
	}
	
#ifdef BNEP_TEST
	if (test_ignore_netfilter == TESTCASE_IGNORE_MESSAGE_ON) {
		BT_DBG("Ignore netfilter\n");
		skb_pull(skb_input, list_length);
		return 0;
	}
#endif

	//Alloc space for new filter
	if (!(new_filter = (struct protocol_range *) 
		kmalloc(sizeof(struct protocol_range[list_length / 4]), GFP_KERNEL))) {
		return -ENOMEM;
	}

	//Copy values
	for (counter = 0; counter < list_length / 4; counter++) {
		memcpy(&new_filter[counter].start, skb_input->data, 2);
		new_filter[counter].start = bnep16_to_cpu(new_filter[counter].start);
		memcpy(&new_filter[counter].stop, &(skb_input->data)[2], 2);
		new_filter[counter].stop = bnep16_to_cpu(new_filter[counter].stop);
		skb_pull(skb_input, 4);

		BT_DBG("filter %d: 0x%.4x - 0x%.4x (%s)", counter,
				new_filter[counter].start,
				new_filter[counter].stop ,
				(new_filter[counter].start <= new_filter[counter].stop) ? "valid" : "invalid");
				
		if (new_filter[counter].start > new_filter[counter].stop) {
			response_code = cpu_to_bnep16(NETFILTER_INVALID_RANGE);
		}		
	}

set_net_filter:

	if (response_code == NETFILTER_OPERATION_SUCCESSFUL) {
		//Delete old filter settings
		if (bnep_conn->filter_info.NetFilter)
			kfree(bnep_conn->filter_info.NetFilter);

		//Set new filter
		bnep_conn->filter_info.NetFilter = new_filter;
		bnep_conn->filter_info.net_filter_count = list_length / 4;
	}

	//Create response control packet
	if (!(skb = alloc_skb(4, GFP_ATOMIC)))
		return -ENOMEM;
	skb_reserve(skb, 2);

	//Insert response code
	memcpy(skb_put(skb, 2), &response_code, 2);

	//Add bnep header and send message    
	bnep_tx_controldata(bnep_conn, BNEP_FILTER_NET_TYPE_RESPONSE_MSG, 0, skb);

	return 0;
}

/***************************************************************************/
static int device_rx_data(struct sk_buff *skb, struct net_device *dev)
{
	struct bnep_private *priv = (struct bnep_private *) dev->priv;
	BT_DBG("Sending packet to higher layer %d", skb->len);

	//Set package options and remove ethernet header
	skb->dev = dev;
	skb->protocol = eth_type_trans(skb, dev);
	skb->ip_summed = CHECKSUM_UNNECESSARY;

	//Update statistics
	priv->stats.rx_packets++;
	priv->stats.rx_bytes += skb->len;

	netif_rx_ni(skb);

	return 0;
}

/***************************************************************************/
static int device_tx_data(struct sk_buff *skb, struct net_device *dev)
{
	if (bnep_dev_counter > 0) {
		struct ethhdr eth;
		struct bnep_private *priv = bnep_get_priv(dev);

		spin_lock_irq(&priv->lock);

		BT_DBG("Sending packet to lower layer %d", skb->len);

		//Update statistics
		priv->stats.tx_packets++;
		priv->stats.tx_bytes += skb->len;

		//Store and remove ethernet header
		memcpy(&eth, skb->data, sizeof(struct ethhdr));
		skb_pull(skb, ETH_HLEN);

		//Forward payload to other slaves
		bnep_tx_etherdata_multiplexor(dev, NULL, skb, &eth, NULL, 0, BNEP_PACKET_TYPE_AUTO);
	
		dev->trans_start = jiffies;

		spin_unlock_irq(&priv->lock);
	}

	return 0;
}

/***************************************************************************/
static int bnep_rx_frame(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{
	//Local var
	__u8 bnep_header;

	//Copy BNEP header
	if (skb_input->len < 1) {
		BT_ERR("packet to short => discard");
		kfree_skb(skb_input);
		return 1;
	}
	
	memcpy(&bnep_header, skb_input->data, 1);
	skb_pull(skb_input, 1);

	BT_DBG("L2CAP Length: 0x%.4x BNEP Type: 0x%.2x Extension: %s",
			skb_input->len + 1, (bnep_header & 0x7F),
			(bnep_header & 0x80) ? "Yes" : "No");

	if ((bnep_header & 0x7F) == BNEP_CONTROL) 
		bnep_rx_controldata(bnep_conn, skb_input, bnep_header & 0x80 );
	else 
		bnep_rx_etherdata(bnep_conn, skb_input, bnep_header & 0x7F, bnep_header & 0x80);
	return 0;
}

/***************************************************************************/
static void bnep_set_multicast_list(kthread_t * kthread)
{
	struct filter_request_struct filter_request;
	struct net_device *dev = NULL;

	BT_DBG("start send multicast thread");

	init_kthread(kthread, "bnep-mc-send");

	while (1) {
		//Wait for changed mc list on any device
		down_interruptible(&bnep_mc_send_sem);
		//Quit thread on terminate signal
		mb();
		if (kthread->terminate) {
			BT_DBG("exiting mc send thread");
			break;
		}
		//Check all devices, if mc list has to be resend
		dev = root_dev;
		while (dev) {
			struct bnep_private *priv = bnep_get_priv(dev);
			struct bnep_connection *bnep_conn;
			struct list_head *ptr;

			if (priv->connection_counter > 0) {
				BT_DBG("dev %s", dev->name);
				//Set general filter settings
				filter_request.filter_type = FILTER_TYPE_MULTICAST;
				filter_request.data = NULL;
				//Set settings for promisc => No filter active
				if (dev->flags & IFF_PROMISC) {
					BT_DBG("promisc => reset multicast filter ");
					filter_request.data_length = 0;
				} else {
					//Set multicast filtering
					bdaddr_t *bdaddr;
					//Prepare multicast filter
					if (dev->flags & IFF_BROADCAST)
						filter_request.data_length =
						    (dev->mc_count + 1) * 2 * sizeof(bdaddr_t);
					else
						filter_request.data_length =
						    (dev->mc_count) * 2 * sizeof(bdaddr_t);
					//Alloc data packet
					if (!(filter_request.data =
					     (void *) kmalloc(filter_request.data_length, GFP_ATOMIC))) {
						BT_ERR(" can't allocate packet for multicast filter");
						return;
					}
					bdaddr = (bdaddr_t *) filter_request.data;
					//Insert broadcast value
					if (dev->flags & IFF_BROADCAST) {
						unsigned char bcast[6] =
						    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
						BT_DBG("multicastlist: Broadcast");
						memcpy(bdaddr++, bcast, sizeof(bdaddr_t));
						memcpy(bdaddr++, bcast, sizeof(bdaddr_t));
					}
					//Add system filter settings
					BT_DBG("multicastlist: add %d entries", dev->mc_count);
					if (dev->mc_count > 0) {
						struct dev_mc_list *mc_ptr = dev->mc_list;
						for (mc_ptr = dev->mc_list; mc_ptr;
						     mc_ptr = mc_ptr->next) {
							BT_DBG("%s", batostr((bdaddr_t *) mc_ptr->dmi_addr));
							//Copy start / stop range
							bacpy(bdaddr++,
							      (bdaddr_t *) mc_ptr->dmi_addr);
							bacpy(bdaddr++,
							      (bdaddr_t *) mc_ptr->dmi_addr);
						}
					}
				}
				//Send settings
				for (ptr = priv->bnep_connections.next;
				     ptr != &priv->bnep_connections; ptr = ptr->next) {
					bnep_conn = list_entry(ptr, struct bnep_connection, list);
					BT_DBG("send settings to node %s",
							batostr(&bluez_pi(bnep_conn->sk->sk)->dst));
					_bnep_tx_filter(bnep_conn, &filter_request);
				}
				//Clean up request
				if (filter_request.data)
					kfree(filter_request.data);
			}
			dev = priv->next_dev;
		}
	}
	//Leave thread
	exit_kthread(kthread);
}

/***************************************************************************/
static void device_set_multicast_list(struct net_device *dev)
{
	struct bnep_private *priv = bnep_get_priv(dev);
	BT_DBG("set new settings");

	//Check, if local node is allowed to set filter
	if (priv->role_uuid != UUID_PANU) {
		BT_DBG("no filter settings allowed on this node due to role GN/NAP");
		return;
	}
	//Send new values to all connections
	up(&bnep_mc_send_sem);
}

/***************************************************************************/
/***************************************************************************/
#ifdef CONFIG_NET_FASTROUTE
static int device_accept_fastpath(struct net_device *dev, struct dst_entry *dst)
{
	return -1;
}
#endif

/***************************************************************************/
/***************************************************************************/
static int bnep_rx_connect_req(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{
	__u8 uuid_length;
	__u16 ResponseMessage = cpu_to_bnep16(CONN_OPERATION_SUCCESSFUL);
	__u16 source_uuid = 0;
	__u16 destination_uuid = 0;
	struct sk_buff *skb;
	int role_uuid = bnep_get_priv(bnep_conn->net_dev)->role_uuid;

	BT_DBG("connection request");

	//Copy & check UUID size 
	if (skb_input->len < 1) goto packet_length_error;
	memcpy(&uuid_length, skb_input->data, 1);
	skb_pull(skb_input, 1);

	//Evaluate UUIDs
	if (skb_input->len < 2 * uuid_length) goto packet_length_error;
	switch (uuid_length) {
	case UUID16:
		memcpy(&destination_uuid, skb_input->data, UUID16);
		destination_uuid = bnep16_to_cpu(destination_uuid);
		memcpy(&source_uuid, &(skb_input->data)[UUID16], UUID16);
		source_uuid = bnep16_to_cpu(source_uuid);
		break;

	case UUID32:	
		memcpy(&destination_uuid , &(skb_input->data)[UUID16] , UUID16);
		destination_uuid = bnep16_to_cpu(destination_uuid);
		memcpy(&source_uuid , &(skb_input->data)[3*UUID16] , UUID16); 						
		source_uuid = bnep16_to_cpu(source_uuid);
		break;
						
	case UUID128:	
		memcpy(&destination_uuid , &(skb_input->data)[UUID16] , UUID16); 
		destination_uuid = bnep16_to_cpu(destination_uuid);
		memcpy(&source_uuid , &(skb_input -> data)[UUID16 + UUID128] , UUID16); 
		source_uuid = bnep16_to_cpu(source_uuid);			
		break;			
	default:
		BT_DBG("illegal uuid_length");
		uuid_length = 0;
		ResponseMessage  = cpu_to_bnep16(CONN_OPERATION_FAILED_INVALID_SERVICE_UUID_SIZE);
	}

	//Remove UUIDs from packet
	skb_pull(skb_input , 2*uuid_length);

	BT_DBG("request uuid_length:0x%x src:0x%x dst:0x%x loc.role:0x%x", uuid_length,
			source_uuid, destination_uuid, role_uuid);

        if (ResponseMessage != cpu_to_bnep16(CONN_OPERATION_FAILED_INVALID_SERVICE_UUID_SIZE)) {
		//Check source_uuid
		if (source_uuid != UUID_PANU && source_uuid != UUID_GN && source_uuid != UUID_NAP)
			ResponseMessage = cpu_to_bnep16(CONN_OPERATION_FAILED_INVALID_SOURCE_UUID);
		//Check source_uuid
		else if (destination_uuid != UUID_PANU && destination_uuid != UUID_GN
			 && destination_uuid != UUID_NAP)
			ResponseMessage = cpu_to_bnep16(CONN_OPERATION_FAILED_INVALID_DESTINATION_UUID);
		//Check connection_uuid
		else {
			if (role_uuid == destination_uuid) {
				if (source_uuid == destination_uuid && source_uuid != UUID_PANU)
					ResponseMessage =
					    cpu_to_bnep16(CONN_OPERATION_FAILED_CONNECTION_NOT_ALLOWED);
				else {
					bnep_conn->destination_uuid = source_uuid;
					bnep_conn -> status = BNEP_CONN_OPEN;
					ResponseMessage = cpu_to_bnep16(CONN_OPERATION_SUCCESSFUL);
				}
			} else
				ResponseMessage = cpu_to_bnep16(CONN_OPERATION_FAILED_INVALID_DESTINATION_UUID);
		}
	}

#ifdef BNEP_TEST
	if (test_ignore_connect == TESTCASE_IGNORE_MESSAGE_ON) {
		BT_DBG("Connection request test cases activated. Reponse will not be sent");
		return 0;
	} 
#endif

	//Send Response packet
	if (!(skb = alloc_skb(4, GFP_ATOMIC)))
		return -ENOMEM;
	skb_reserve(skb, 2);

	//Insert connection response data
	BT_DBG("Send connection response packet");
	memcpy(skb_put(skb, 2), &ResponseMessage, 2);
	bnep_tx_controldata(bnep_conn, BNEP_SETUP_CONNECTION_RESPONSE_MSG, 0, skb);
	return 0;

packet_length_error:
	BT_ERR("connect req packet to short => discard");
	return 1;

}

/***************************************************************************/
static int bnep_rx_connect_rsp(struct bnep_connection *bnep_conn, struct sk_buff *skb_input)
{

	__u16 ResponseMessage;

	//Process connection request data
	if (skb_input->len < 2) {
		BT_ERR("packet to short => discard");
		return 1;
	}
	
	memcpy(&ResponseMessage, skb_input->data, 2);
	ResponseMessage = bnep16_to_cpu(ResponseMessage);
	skb_pull(skb_input, 2);

	BT_DBG("Got connection response : 0x%x", ResponseMessage);

	if (ResponseMessage == CONN_OPERATION_SUCCESSFUL) {
		BT_DBG("open now connection for communication\n");
		bnep_conn -> status = BNEP_CONN_OPEN;
	}
	
	if (bnep_conn->connect_request.req_status == BNEP_REQ_PENDING) {
		bnep_conn->connect_request.req_status = BNEP_REQ_DONE;
		bnep_conn->connect_request.req_result = ResponseMessage;
		wake_up_interruptible(&(bnep_conn->connect_request.req_wait_q));
	} else {
		BT_DBG("connection not waiting for connection response => discard");
	}
	
	return 0;
}

/***************************************************************************/
static void bnep_filter_req(struct request_struct *request, unsigned long opt)
{
	struct filter_request_struct *filter_request = (struct filter_request_struct *) opt;

	if (request->bnep_conn) {
		struct sk_buff *skb;
		__u16 list_size;
		__u8 cmd_type;

		//Eval request dep values for multicast/protocol packet
		switch (filter_request->filter_type) {
		case FILTER_TYPE_PROTOCOL:
			cmd_type = BNEP_FILTER_NET_TYPE_SET_MSG;
			break;
		case FILTER_TYPE_MULTICAST:
			cmd_type = BNEP_FILTER_MULT_ADDR_SET_MSG;
			break;
		default:
			BT_DBG("filter request with unknown type");
			return;
		}

		if (!(skb = alloc_skb(2 + 2 + filter_request->data_length, GFP_ATOMIC))) {
			BT_ERR("can't alloc space for filter packet");
			return;
		}
		//Reserve headspace for control packet
		skb_reserve(skb, 2);

		//List size = Number of bytes in request
		list_size = cpu_to_bnep16(filter_request->data_length);
		memcpy(skb_put(skb, 2), &list_size, 2);

		//Copy request values
		if (filter_request->data_length > 0) {
			memcpy(skb_put(skb, filter_request->data_length), filter_request->data,
			       filter_request->data_length);
		}
		//complete and send control package
		bnep_tx_controldata(request->bnep_conn, cmd_type, 0, skb);
	}
}

/***************************************************************************/
int _bnep_tx_filter(struct bnep_connection *bnep_conn, struct filter_request_struct *filter_request)
{
	int resend_counter = 0;
	int status = 0;

	if (!bnep_conn) {
		BT_DBG("not l2cap connected");
		return BNEP_ERR_NO_L2CAP;
	}

	if (bnep_conn->filter_request.req_status != BNEP_REQ_DONE) {
		BT_DBG("pending filter request found on this connection");
		return BNEP_ERR_BNEP_FILTER_REQ_PENDING;
	}

	do {
		BT_DBG("issue filter request #%d", resend_counter);
		bnep_conn->filter_request.req_status = BNEP_REQ_PENDING;
		bnep_conn->filter_request.bnep_conn = bnep_conn;
		bnep_request(&bnep_conn->filter_request, bnep_filter_req,
			     (unsigned long) filter_request, BNEP_FILTER_TIMEOUT * HZ);
		resend_counter++;
	} while (bnep_conn->filter_request.req_status != BNEP_REQ_DONE && resend_counter < 3);

	if (bnep_conn->filter_request.req_status != BNEP_REQ_DONE)
		status = BNEP_ERR_FILTER_TIMEOUT;
	else {
		if (bnep_conn->filter_request.req_result != 0x0)
			status = BNEP_ERR_FILTER_ERROR;
	}

	bnep_conn->filter_request.req_status = BNEP_REQ_DONE;

	return status;

}

/***************************************************************************/
int bnep_tx_filter(struct net_device *net_dev, struct filter_request_struct *filter_request)
{
	struct bnep_connection *bnep_conn =
	    bnep_get_conn_by_bdaddr(net_dev, &filter_request->destination_address);
	return _bnep_tx_filter(bnep_conn, filter_request);
}

/***************************************************************************/
static void bnep_start_session(struct socket *sock, struct net_device *net_dev)
{
	struct bnep_connection *bnep_conn = NULL;
	struct bnep_private *priv = bnep_get_priv(net_dev);

	BT_DBG("L2CAP connection for BNEP open");

	if ((bnep_conn =
	     (struct bnep_connection *) kmalloc(sizeof(struct bnep_connection), GFP_KERNEL))) {
		BT_DBG("adding connection list element");
		bnep_conn -> status = BNEP_CONN_CLOSED;
		bnep_conn->sk = sock;
		bnep_conn->net_dev = net_dev;

		//Init filter req data
		bnep_conn->filter_info.NetFilter = NULL;
		bnep_conn->filter_info.MulticastFilter = NULL;

		//Init connect req struct
		init_waitqueue_head(&(bnep_conn->connect_request.req_wait_q));
		bnep_conn->connect_request.req_status = BNEP_REQ_DONE;
		init_MUTEX(&(bnep_conn->connect_request.req_lock));

		//Init filter req struct        
		init_waitqueue_head(&(bnep_conn->filter_request.req_wait_q));
		bnep_conn->filter_request.req_status = BNEP_REQ_DONE;
		init_MUTEX(&(bnep_conn->filter_request.req_lock));

		//Register new connection
		list_add(&bnep_conn->list, &priv->bnep_connections);
		priv->connection_counter++;
		MOD_INC_USE_COUNT;

		//Prepare and start session receive thread
		bnep_conn->session.arg = (void *) bnep_conn;
		start_kthread(bnep_session, &bnep_conn->session);
	} else {
		BT_ERR("can't allocate mem for new connection");
	}
}

/***************************************************************************/
static void bnep_tx_frame(struct socket *sk, struct sk_buff *skb)
{
	mm_segment_t old_fs;
	struct msghdr msg;
	struct iovec iv;
	int status;

	iv.iov_base = (char *) skb->data;
	iv.iov_len = (size_t) skb->len;
	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &iv;
	msg.msg_iovlen = 1;
	msg.msg_flags = MSG_NOSIGNAL;

	old_fs = get_fs();
	set_fs(KERNEL_DS);
	status = sock_sendmsg(sk, &msg, (size_t) skb->len);
	BT_DBG("%d bytes send", status);
	set_fs(old_fs);

	dev_kfree_skb(skb);
}

/***************************************************************************/
static void bnep_session(kthread_t * kthread)
{
	struct list_head *ptr;
	struct bnep_connection *entry = (struct bnep_connection *) kthread->arg;
	struct socket *session_sk = entry->sk;
	struct bnep_private *priv = bnep_get_priv(entry->net_dev);
	wait_queue_t wait;

	BT_DBG("start session thread");

	init_kthread(kthread, "bnep_rcv");

	init_waitqueue_entry(&wait, current);
	add_wait_queue(session_sk->sk->sleep, &wait);
	while (1) {
		struct sk_buff *skb;

		set_current_state(TASK_INTERRUPTIBLE);

		//Close connection
		if (session_sk->sk->state == BT_CLOSED || kthread->terminate) {
			break;
		}

		//Get all data from L2CAP
		while ((skb = skb_dequeue(&session_sk->sk->receive_queue))) {
			//Send skb to bnep
			skb_orphan(skb);
			bnep_rx_frame(entry, skb);
		}

		//Send all data to L2CAP
		while ((skb = skb_dequeue(&session_sk->sk->write_queue))) {
			bnep_tx_frame(session_sk, skb);
		}

		schedule();
	}
	remove_wait_queue(session_sk->sk->sleep, &wait);

	for (ptr = priv->bnep_connections.next; ptr != &priv->bnep_connections; ptr = ptr->next) {
		entry = list_entry(ptr, struct bnep_connection, list);
		if (entry->sk == session_sk) {
			BT_DBG("deleting connection list entry");
			--(priv->connection_counter);
			MOD_DEC_USE_COUNT;
			list_del(&entry->list);
			if (entry->filter_info.NetFilter)
				kfree(entry->filter_info.NetFilter);
			if (entry->filter_info.MulticastFilter)
				kfree(entry->filter_info.MulticastFilter);
			break;
		}
	}

	BT_DBG("releasing socket");
	sock_release(session_sk);

	//Leave thread
	exit_kthread(kthread);
}

/***************************************************************************/
static void bnep_connect_req(struct request_struct *request, unsigned long opt)
{
	struct bnep_connect_mgnt_struct *bnep_connect_mgnt =
	    (struct bnep_connect_mgnt_struct *) opt;
	int role_uuid = bnep_get_priv(request->bnep_conn->net_dev)->role_uuid;

	if (request->bnep_conn) {
		//unsigned long data = 0;
		int packet_size;
		struct sk_buff *skb;

		//Connect parameter
		__u16 source_uuid = cpu_to_bnep16(role_uuid);
		__u16 dest_uuid = cpu_to_bnep16(bnep_connect_mgnt->destination_uuid);
		__u8 uuid_length = UUID16;

		//Prepare skb
		packet_size = 2 + 1 + 2 * uuid_length;
		if (!(skb = alloc_skb(packet_size, GFP_ATOMIC))) {
			request->req_result = -ENOMEM;
			return;
		}
		skb_reserve(skb, packet_size);

		//Insert connection request data
		memcpy(skb_push(skb, uuid_length), &source_uuid, uuid_length);
		memcpy(skb_push(skb, uuid_length), &dest_uuid, uuid_length);
		memcpy(skb_push(skb, 1), &uuid_length, 1);

		//complete and send control package
		bnep_tx_controldata(request->bnep_conn, BNEP_SETUP_CONNECTION_REQUEST_MSG, 0, skb);
	}
	return;
}

/***************************************************************************/
static int bnep_l2cap_connect(struct net_device *net_dev,
			      struct bnep_connect_mgnt_struct *bnep_connect_mgnt)
{
	struct l2cap_options opts;
	struct sockaddr_l2 addr;
	struct socket *nsk;
	int err, size;
	struct bnep_private *bnep_priv = bnep_get_priv(net_dev);

	if (bnep_get_conn_by_bdaddr(net_dev, &bnep_connect_mgnt->destination_address)) {
		BT_DBG("L2CAP connection on PAN PSM to addr %s already exists",
				batostr(&bnep_connect_mgnt->destination_address));
		return 0;
	}

	err = sock_create(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP, &nsk);
	if (nsk) {
		BT_DBG("new socket accepted %p", nsk);
	} else {
		BT_DBG("new socket not accepted");
		return BNEP_ERR_SOCK_ACCEPT;
	}

	memset(&addr, 0, sizeof(struct sockaddr_l2));
	addr.l2_family = AF_BLUETOOTH;
	baswap(&addr.l2_bdaddr, (bdaddr_t *) & bnep_priv->source_address);
	err = nsk->ops->bind(nsk, (struct sockaddr *) &addr, sizeof(addr));
	if (err < 0) {
		BT_DBG("bind failed %d", err);
		return BNEP_ERR_SOCK_BIND;
	}
	//Set L2CAP options 
	size = sizeof(opts);
	nsk->ops->getsockopt(nsk, SOL_L2CAP, L2CAP_OPTIONS, (void *) &opts, &size);
	opts.imtu = PAN_IMTU;
	opts.omtu = PAN_OMTU;
	opts.flush_to = PAN_FLUSH_TO;

	set_fs(KERNEL_DS);
	nsk->ops->setsockopt(nsk, SOL_L2CAP, L2CAP_OPTIONS, (void *) &opts, size);

	addr.l2_psm = htobs(PAN_PSM);
	baswap(&addr.l2_bdaddr, &bnep_connect_mgnt->destination_address);
	if (!nsk->ops->connect(nsk, (struct sockaddr *) &addr, sizeof(struct sockaddr), 0)) {
		BT_DBG("Connection ready");
		bnep_start_session(nsk, net_dev);
	} else {
		BT_DBG("Can't connect %s -> %s", batostr(&bnep_priv->source_address),
				batostr(&bnep_connect_mgnt->destination_address));
		sock_release(nsk);
		return BNEP_ERR_CONNECT_L2CAP;
	}

	return 0;
}

/***************************************************************************/
static int bnep_connect(struct net_device *net_dev,
			struct bnep_connect_mgnt_struct *bnep_connect_mgnt)
{
	struct bnep_connection *bnep_conn =
	    bnep_get_conn_by_bdaddr(net_dev, &bnep_connect_mgnt->destination_address);
	int status = 0;
	int resend_counter = 0;
	int role_uuid = bnep_get_priv(net_dev)->role_uuid;

	if (role_uuid != UUID_PANU && role_uuid != UUID_GN && role_uuid != UUID_NAP)
		return BNEP_ERR_CONN_UNKNOWN_DEF_ROLE;

	if (!bnep_conn) {
		BT_DBG("l2cap connection not ready => start connection");
		if (bnep_l2cap_connect(net_dev, bnep_connect_mgnt) != 0 ||
			!(bnep_conn = bnep_get_conn_by_bdaddr(net_dev,
						&bnep_connect_mgnt->destination_address))) {
			BT_DBG("could not establish l2cap connection");
			return BNEP_ERR_NO_L2CAP;
		}
		BT_DBG("l2cap connection now ready");
	}

	if (bnep_conn->connect_request.req_status != BNEP_REQ_DONE) {
		BT_DBG("pending bnep connection request found on this connection");
		return BNEP_ERR_BNEP_CONN_REQ_PENDING;
	}

	do {
		BT_DBG("issue bnep connect request #%d", resend_counter);
		bnep_conn->connect_request.req_status = BNEP_REQ_PENDING;
		bnep_conn->connect_request.bnep_conn = bnep_conn;
		bnep_request(&bnep_conn->connect_request, bnep_connect_req,
			     (unsigned long) bnep_connect_mgnt, BNEP_CONN_TIMEOUT * HZ);
		resend_counter++;
	} while (bnep_conn->connect_request.req_status != BNEP_REQ_DONE && resend_counter < 3);

	if (bnep_conn->connect_request.req_status != BNEP_REQ_DONE)
		status = BNEP_ERR_CONNECT_BNEP_TIMEOUT;
	else {
		if (bnep_conn->connect_request.req_result != CONN_OPERATION_SUCCESSFUL)
			status = BNEP_ERR_CONNECT_BNEP_ERROR;
		else
			bnep_conn->destination_uuid = bnep_connect_mgnt->destination_uuid;
	}

	bnep_conn->connect_request.req_status = BNEP_REQ_DONE;

	net_dev->set_multicast_list(net_dev);

	return status;
}

/***************************************************************************/
static int _bnep_request(struct request_struct *request,
	 void (*req) (struct request_struct *, unsigned long), unsigned long opt, __u32 timeout)
{
	DECLARE_WAITQUEUE(wait, current);
	int err = 0;

	add_wait_queue(&request->req_wait_q, &wait);
	current->state = TASK_INTERRUPTIBLE;

	req(request, opt);
	schedule_timeout(timeout);

	current->state = TASK_RUNNING;
	remove_wait_queue(&request->req_wait_q, &wait);

	if (signal_pending(current))
		return -EINTR;

	switch (request->req_status) {
	case BNEP_REQ_DONE:
		err = 0;
		break;
	case BNEP_REQ_CANCELED:
		err = -request->req_result;
		break;

	default:
		err = -ETIMEDOUT;
		break;
	};

	BT_DBG("end: err %d", err);

	return err;
}

/***************************************************************************/
static inline int bnep_request(struct request_struct *request,
				void (*req) (struct request_struct *, unsigned long),
				unsigned long opt, __u32 timeout)
{
	int ret;
	//Serialize all requests
	down(&request->req_lock);
	init_waitqueue_head(&request->req_wait_q);
	ret = _bnep_request(request, req, opt, timeout);
	up(&request->req_lock);
	return ret;
}

/***************************************************************************/
static int bnep_update_role(struct net_device *dev, int new_role)
{
	struct list_head *ptr;
	struct bnep_connection *entry;
	struct bnep_connect_mgnt_struct bnep_connect_mgnt;
	bdaddr_t tmp_addr;
	struct bnep_private *priv = bnep_get_priv(dev);

	if (new_role != UUID_PANU && new_role != UUID_GN && new_role != UUID_NAP)
		return BNEP_ERR_CONN_UNKNOWN_DEF_ROLE;

	priv->role_uuid = new_role;
	BT_DBG("set role to 0x%x", new_role);

	//Send updated BNEP connect message to all connected nodes
	for (ptr = priv->bnep_connections.next; ptr != &priv->bnep_connections; ptr = ptr->next) {
		entry = list_entry(ptr, struct bnep_connection, list);
		baswap(&tmp_addr, &bluez_pi(entry->sk->sk)->dst);
		//prepare updated bnep connect request
		bacpy(&bnep_connect_mgnt.destination_address, &tmp_addr);
		bnep_connect_mgnt.protocol = CONN_BNEP_LAYER;
		bnep_connect_mgnt.destination_uuid = entry->destination_uuid;
		//send updated connect request
		bnep_connect(dev, &bnep_connect_mgnt);
	}

	return 0;
}

/***************************************************************************/
static int bnep_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	switch (cmd) {
	//Close a connection            
	case SIOC_DISCONNECT:{
			struct bnep_connect_mgnt_struct bnep_connect_mgnt;
			struct bnep_connection *conn = NULL;
			copy_from_user(&bnep_connect_mgnt, rq->ifr_data,
				       sizeof(struct bnep_connect_mgnt_struct));
			conn = bnep_get_conn_by_bdaddr(dev, &bnep_connect_mgnt.destination_address);
			if (conn) {
				BT_DBG("closing connection to %s",
						batostr(&bnep_connect_mgnt.destination_address));

				//Kill session thread
				stop_kthread(&conn->session);
			} else {
				BT_DBG("connection to %s not found",
						batostr(&bnep_connect_mgnt.destination_address));
				return BNEP_ERR_NO_L2CAP;
			}
			return 0;
		}

		//Create a BNEP or L2CAP connection             
	case SIOC_CONNECT:{
			int status = 0;
			struct bnep_connect_mgnt_struct bnep_connect_mgnt;

			copy_from_user(&bnep_connect_mgnt, rq->ifr_data,
				       sizeof(struct bnep_connect_mgnt_struct));
			switch (bnep_connect_mgnt.protocol) {
			case CONN_L2CAP_LAYER:
				status = bnep_l2cap_connect(dev, &bnep_connect_mgnt);
				break;
			case CONN_BNEP_LAYER:
				status = bnep_connect(dev, &bnep_connect_mgnt);
				break;
			default:
				status = BNEP_ERR_CONN_UNKNOWN_LAYER;
			}
			return status;
		}

		//Set filter on specific link
	case SIOC_SET_FILTER:{
			struct filter_request_struct filter_request;
			unsigned char *data = NULL;
			int status;
			copy_from_user(&filter_request, rq->ifr_data,
				       sizeof(struct filter_request_struct));
			if (filter_request.data_length > 0) {
				data =
				    (unsigned char *) kmalloc(filter_request.data_length,
							      GFP_ATOMIC);
				copy_from_user(data, filter_request.data,
					       filter_request.data_length);
				filter_request.data = data;
			}
			status = bnep_tx_filter(dev, &filter_request);
			kfree(data);
			return status;
		}

		//Change role of device
	case SIOC_SET_ROLE:{
			int role_uuid;
			copy_from_user(&role_uuid, rq->ifr_data, sizeof(int));
			return bnep_update_role(dev, role_uuid);
		}

		//Test case execution
	case SIOC_TESTCASE:	/* Execute test case  */  {
#ifdef BNEP_TEST
			struct bnep_test_ctrl_struct bnep_test_ctrl;
			copy_from_user(&bnep_test_ctrl, rq->ifr_data,
				       sizeof(struct bnep_test_ctrl_struct));
			return test_execute_test(dev, &bnep_test_ctrl);
#else
			BT_DBG("test mode not available. Recompile with -DBNEP_TEST");
			return 1;
#endif
			break;
		}
	}

	return -EOPNOTSUPP;
}

/***************************************************************************/
static struct net_device_stats *device_get_stats(struct net_device *dev)
{
	struct bnep_private *priv;
	priv = (struct bnep_private *) dev->priv;
	return &(priv->stats);
}

/***************************************************************************/
static int device_open_dev(struct net_device *dev)
{
	netif_start_queue(dev);
	return 0;
}

/***************************************************************************/
static int device_close_dev(struct net_device *dev)
{
	netif_stop_queue(dev);
	return 0;
}

/***************************************************************************/
static void device_dev_timeout(struct net_device *dev)
{
	BT_DBG("not implemented");
}

/***************************************************************************/
static int __init bnep_init(struct net_device *dev)
{
	int i;
	struct bnep_private *priv = (struct bnep_private *) dev->priv;

	//init dev struct
	ether_setup(dev);

	//net device setup
	dev->hard_start_xmit = &device_tx_data;
	dev->get_stats = &device_get_stats;
	dev->do_ioctl = &bnep_ioctl;
	dev->set_multicast_list = &device_set_multicast_list;
	dev->open = &device_open_dev;
	dev->stop = &device_close_dev;
	dev->tx_timeout = &device_dev_timeout;
	dev->watchdog_timeo = HZ;
#ifdef CONFIG_NET_FASTROUTE
	dev->accept_fastpath = device_accept_fastpath;
#endif

	//Set device source address from private data
	BT_DBG("Binding %s <-> %s", batostr(&priv->source_address), dev->name);
	for (i = 0; i < 6; i++)
		dev->dev_addr[i] = ((unsigned char *) &priv->source_address)[i];

	dev->tx_queue_len = 100;
	dev->flags &= (IFF_MULTICAST | IFF_BROADCAST);

	return 0;
}

/***************************************************************************/
struct socket *bnep_sock_accept(struct socket *sk)
{
	struct socket *nsk;
	int err;

	nsk = sock_alloc();
	if (!nsk)
		return NULL;

	nsk->type = sk->type;
	nsk->ops = sk->ops;
	err = sk->ops->accept(sk, nsk, 0);
	BT_DBG("result: %d", err);

	if (err < 0) {
		sock_release(nsk);
		return NULL;
	}

	return nsk;
}

/***************************************************************************/
static void bnep_connect_listener(kthread_t * kthread)
{
	struct sockaddr_l2 addr;
	struct l2cap_options opts;
	int err, size;
	struct net_device *dev = (struct net_device *) kthread->arg;
	struct bnep_private *bnep_priv = bnep_get_priv(dev);

	BT_DBG("start connect listener");

	init_kthread(kthread, "bnep_conn_lsten");

	set_fs(KERNEL_DS);

	// Create socket
	err = sock_create(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP, &bnep_priv->bnep_sk);
	if (err < 0) {
		BT_DBG("Create socket failed %d", err);
		return;
	}
	// Bind socket 
	baswap(&addr.l2_bdaddr, &bnep_priv->source_address);
	addr.l2_family = AF_BLUETOOTH;
	addr.l2_psm = htobs(PAN_PSM);
	err =
	    bnep_priv->bnep_sk->ops->bind(bnep_priv->bnep_sk, (struct sockaddr *) &addr,
					  sizeof(addr));
	if (err < 0) {
		BT_DBG("Bind failed %d", err);
		return;
	}
	// Set L2CAP options
	size = sizeof(opts);
	bnep_priv->bnep_sk->ops->getsockopt(bnep_priv->bnep_sk, SOL_L2CAP, L2CAP_OPTIONS,
					    (void *) &opts, &size);
	opts.imtu = PAN_IMTU;
	opts.imtu = PAN_OMTU;
	opts.flush_to = PAN_FLUSH_TO;
	bnep_priv->bnep_sk->ops->setsockopt(bnep_priv->bnep_sk, SOL_L2CAP, L2CAP_OPTIONS,
					    (void *) &opts, size);

	// Start listening on the socket
	err = bnep_priv->bnep_sk->ops->listen(bnep_priv->bnep_sk, 10);
	if (err) {
		BT_DBG("Listen failed %d", err);
		sock_release(bnep_priv->bnep_sk);
		return;
	}

	while (1) {
		struct socket *nsk;

		mb();
		if (kthread->terminate) {
			BT_DBG("exiting connect listener");
			break;
		}

		nsk = bnep_sock_accept(bnep_priv->bnep_sk);
		if (nsk) {
			bdaddr_t tmp_addr;
			baswap(&tmp_addr, &bluez_pi(nsk->sk)->dst);

			//Check for existing bnep connection
			if (bnep_get_conn_by_bdaddr(dev, &tmp_addr)) {
				BT_DBG("connection to %s exists => close new connection",
						batostr(&tmp_addr));
				sock_release(nsk);
			} else {
				//Start new thread for this session
				BT_DBG("starting bnep session on sock %p", nsk);
				bnep_start_session(nsk, (struct net_device *) kthread->arg);
			}
		}
	}

	//Kill thread
	exit_kthread(kthread);
	return;
}

/***************************************************************************/
static int bnep_read_proc(char *buf, char **start, off_t offset, int count, int *eof, void *priv)
{
	char *ptr = buf;
	int len;
	struct list_head *list_ptr;
	struct bnep_connection *bnep_conn;
	struct net_device *net_dev = root_dev;
	int counter;
	bdaddr_t tmp_addr;
	struct bnep_private *bnep_priv;

	ptr += sprintf(ptr, "BNEP device status:\n");
	if (net_dev == NULL) {
		ptr += sprintf(ptr, " No PAN devices loaded\n");
	} else {
		while (net_dev) {
			bnep_priv = bnep_get_priv(net_dev);
			ptr += sprintf(ptr, " %s\tBDAddr: %s Role: 0x%.4x\n", net_dev->name,
				    batostr(&bnep_priv->source_address), bnep_priv->role_uuid);

			for (list_ptr = bnep_priv->bnep_connections.next;
			     list_ptr != &bnep_priv->bnep_connections; list_ptr = list_ptr->next) {
				bnep_conn = list_entry(list_ptr, struct bnep_connection, list);
				baswap(&tmp_addr, &bluez_pi(bnep_conn->sk->sk)->dst);
				ptr += sprintf(ptr, "\t  Connection: %s Role 0x%.4x\n",
					    batostr(&tmp_addr), bnep_conn->destination_uuid);

				if (bnep_conn->filter_info.MulticastFilter) {
					for (counter = 0;
					     counter <
					     bnep_conn->filter_info.multicast_filter_count;
					     counter++) {
						int byte_counter;
						ptr +=
						    sprintf(ptr, "\t\tMulticast Filter %d: ",
							    counter);
						for (byte_counter = 5; byte_counter >= 0;
						     byte_counter--) {
							ptr +=
							    sprintf(ptr, "%.2x%s",
								    ((__u8 *) & bnep_conn->
								     filter_info.
								     MulticastFilter[counter].
								     start)[byte_counter],
								    byte_counter == 0 ? "" : ":");
						}
						ptr += sprintf(ptr, " - ");
						for (byte_counter = 5; byte_counter >= 0;
						     byte_counter--) {
							ptr +=
							    sprintf(ptr, "%.2x%s",
								    ((__u8 *) & bnep_conn->
								     filter_info.
								     MulticastFilter[counter].
								     stop)[byte_counter],
								    byte_counter == 0 ? "" : ":");
						}
						ptr += sprintf(ptr, "\n");
					}
				} else {
					ptr += sprintf(ptr, "\t\tNo Multicast Filter active\n");
				}

				if (bnep_conn->filter_info.NetFilter) {
					for (counter = 0;
					     counter < bnep_conn->filter_info.net_filter_count;
					     counter++) {
						ptr +=
						    sprintf(ptr,
							    "\t\tProtocol Filter %d: 0x%.4x - 0x%.4x\n",
							    counter,
							    bnep_conn->filter_info.
							    NetFilter[counter].start,
							    bnep_conn->filter_info.
							    NetFilter[counter].stop);
					}
				} else {
					ptr += sprintf(ptr, "\t\tNo Prototocl Filter active\n");
				}
			}
			net_dev = bnep_priv->next_dev;
			ptr += sprintf(ptr, "\n");
		}
	}

	len = ptr - buf;

	if (len <= count + offset)
		*eof = 1;

	*start = buf + offset;
	len -= offset;

	if (len > count)
		len = count;
	if (len < 0)
		len = 0;

	return len;
}

/******************************************************************************/
int bnep_dev_control(struct bnep_dev_ioctl_struct *bnep_dev_ioctl)
{
	int err = 0;

	if (bnep_dev_ioctl->net_device_command == NET_DEVICE_REMOVE) {

		//Find device in device list
		struct net_device *del_dev = NULL;
		struct net_device *pre_dev = NULL;
		struct net_device *tmp_dev = root_dev;
		struct bnep_private *priv = NULL;
		int status;

		//Remove device from dev list
		while (tmp_dev && !del_dev) {
			if (strcmp(tmp_dev->name, bnep_dev_ioctl->name) == 0) {
				//Check for active connections
				if (bnep_get_priv(tmp_dev)->connection_counter != 0) {
					BT_DBG("can't remove device with active connections");
					return BNEP_ERR_DEVICE_CONNECTED;
				}
				del_dev = tmp_dev;
				if (pre_dev)
					bnep_get_priv(pre_dev)->next_dev =
					    bnep_get_priv(tmp_dev)->next_dev;
				else
					root_dev = bnep_get_priv(tmp_dev)->next_dev;
			} else {
				pre_dev = tmp_dev;
				tmp_dev = bnep_get_priv(tmp_dev)->next_dev;
			}
		}

		if (!del_dev) {
			BT_DBG("device not loaded");
			return BNEP_ERR_DEV_NOT_LOADED;
		}

		BT_DBG("remove eth net device");

		priv = bnep_get_priv(del_dev);

		//Stop connection listener thread
		stop_kthread(&priv->listener_thread);

		//Clean interface to l2cap
		if (priv->bnep_sk)
			sock_release(priv->bnep_sk);

		//Clean interface from kernel
		unregister_netdev(del_dev);
		status = netdev_finish_unregister(del_dev);
		if (status)
			BT_DBG("can't free net device mem");

		bnep_dev_counter--;
		MOD_DEC_USE_COUNT;
	} else if (bnep_dev_ioctl->net_device_command == NET_DEVICE_INSERT) {
		struct bnep_private *priv;
		struct net_device *new_dev;
		BT_DBG("insert eth net device");

		//Allocate new eth device
		if (!(new_dev = alloc_etherdev(sizeof(struct bnep_private))))
			return -ENOMEM;

		// Find a name for this unit 
		if (dev_alloc_name(new_dev, bnep_dev_ioctl->name) < 0) {
			kfree(new_dev);
			return BNEP_ERR_ALLOC_NAME;
		}
		//Init device
		priv = bnep_get_priv(new_dev);
		priv->bnep_sk = NULL;
		priv->role_uuid = bnep_dev_ioctl->role_uuid;
		new_dev->init = bnep_init;
		spin_lock_init(&priv->lock);
		bacpy(&priv->source_address, &bnep_dev_ioctl->source_address);

		//Init list for bnep connections
		INIT_LIST_HEAD(&priv->bnep_connections);
		priv->connection_counter = 0;

		//Put new device in device chain
		priv->next_dev = root_dev;
		root_dev = new_dev;

		//Register new device in linux
		BT_DBG("register dev to linux");
		if (register_netdev(new_dev) != 0)
			return BNEP_ERR_REGISTER_NETDEV;

		//Init connect listener
		BT_DBG("init connect listener");
		priv->listener_thread.arg = (void *) new_dev;
		start_kthread(bnep_connect_listener, &priv->listener_thread);

		MOD_INC_USE_COUNT;
		bnep_dev_counter++;
	} else {
		BT_DBG("netdevice command not understood");
		err = BNEP_ERR_UNKNOWN_NETDEV_COMMAND;
	}

	return err;
}

/***************************************************************************/
static int __init bnep_init_module(void)
{
	bnep_dev_counter = 0;
	root_dev = NULL;
	create_proc_read_entry("net/bnep", 0, 0, bnep_read_proc, NULL);

	BT_INFO("BNEP ver 1.0.4  Copyright (C) 2001 Sony Corporation");
	BT_INFO("Written 2001,2002 by Johannes Loebbert <loebbert@sony.de>");

	//Used only for mod counting
	SET_MODULE_OWNER(&bnep_owner);

	sema_init(&bnep_mc_send_sem, 0);
	start_kthread(bnep_set_multicast_list, &bnep_mc_send_thread);

	//Register BNEP_Sock Layer for dev control
	bnep_sock_init();

#ifdef BNEP_TEST
	//Prepare custom test packet          
	test_bnep_packet_prepare();
#endif
	return 0;
}

/***************************************************************************/
static void __exit bnep_cleanup_module(void)
{
	struct bnep_dev_ioctl_struct bnep_dev_ioctl;

	//Deregister BNEP_Sock
	bnep_sock_cleanup();

	stop_kthread(&bnep_mc_send_thread);

	//Remove all active devices
	bnep_dev_ioctl.net_device_command = NET_DEVICE_REMOVE;
	while (root_dev) {
		BT_DBG("Remove device %s", root_dev->name);
		strcpy(bnep_dev_ioctl.name, root_dev->name);
		bnep_dev_control(&bnep_dev_ioctl);
	}

	remove_proc_entry("net/bnep", NULL);
}

module_init(bnep_init_module);
module_exit(bnep_cleanup_module);

MODULE_AUTHOR("Johannes Loebbert <loebbert@sony.de>");
MODULE_DESCRIPTION("BNEP ver 1.0.4");
MODULE_LICENSE("GPL");
